/**
 * Communication with the
 * system through the Jolicloud daemon
 */
var JolicloudDaemon = {
	/**
	 * Calls a function of the daemon
	 * session: the session to send the result
	 * callId: the id of the call to reply to
	 * object: path to the object to call
	 * functionName: name of the function
	 * args: arguments to send to the function
	 * formatter: function to format the result if success
	 */
	callAndSend: function(session, callId, object, functionName, args, formatter)
	{
		if (!args) 
			args = [];
		//we prepare the method
		var callbackObject = {
			'object': this,
			'session': session,
			'callId': callId,
			'formatter': formatter
		}
		var method = dbus.getMethod(dbus.SESSION, "org.jolicloud.JolicloudDaemon", object, functionName, "org.jolicloud.JolicloudDaemon", null, callbackObject);
		method.onreply = this._onReplyCallback;
		method.onerror = this._onErrorCallback;
		//Very important reference:
		// prevents destruction of the method: callback would never be called
		callbackObject.method = method;
		method.apply(method, args);
	},
	
	/**
	 * Listen to events of the daemon
	 * session: the session to send the result
	 * callId: the id of the call to send events to
	 * listenerId: id of the listener
	 * activate: start listening if true, stop if false
	 * object: path to the object to call
	 * signalName: name of the signal
	 * formatter: function to format the result if success
	 */
	listenAndSend: function(session, callId, listenerId, activate, object, signalName, formatter)
	{
	    if (!activate)
		{
			//Deactivate the listener
			if (session.externalDestroyers[listenerId])
				session.externalDestroyers[listenerId]();
			delete session.externalDestroyers[listenerId];
		}
		else
		{
			//Activate the listener
			//we prepare the method
			var callbackObject = {
				'object': this,
				'session': session,
				'callId': callId,
				'formatter': formatter
			}
			var signal = dbus.getSignal(dbus.SESSION, "org.jolicloud.JolicloudDaemon", signalName, null, object, callbackObject);
			signal.onemit = this._onSignalCallback;
			//Very important reference:
			// prevents destruction of the signal: callback would never be called
			callbackObject.signal = signal;
			signal.enabled = true;
			//we save the stop function to use it later
			if (session.externalDestroyers[listenerId])
				session.externalDestroyers[listenerId]();
			session.externalDestroyers[listenerId] = function()
			{
				signal.enabled = false;
				signal = null;
			}
		}
	},
	
	/**
	 * Process a reply
	 * @param {Object} result
	 */
	_onReplyCallback: function(result)
	{
		result = this.object._formatResponse(result);
		if (this.formatter) 
			result = this.formatter(result);
		this.session.callbackCall(this.callId, result);
	},
	
	/**
	 * Process an error
	 * @param {Object} error
	 */
	_onErrorCallback: function(error)
	{
		if (error.substr(0, 20) == "org.jolicloud.Error.") 
			this.session.callbackError(this.callId, error.substr(20));
		else 
		{
			this.session.log("DBUS error: " + error);
			this.session.callbackError(this.callId, "JOLICLOUD_DAEMON_NOT_AVAILABLE");
		}
	},
	
	/**
	 * Process a signal
	 */
	_onSignalCallback: function()
	{
		args = [];
		for (var i = 0; i < arguments.length; i++) 
			args.push(arguments[i]);
		result = this.object._formatResponse(args);
		if (this.formatter) 
			result = this.formatter(result);
		this.session.callbackEvent(this.callId, result);
	},
	
	/**
	 * Format a DBUS response, replace '' by null
	 * @param {Object} object
	 */
	_formatResponse: function(object)
	{
		var new_object = null;
		if (object != null && typeof(object) == 'object') 
		{
			if (object.length && object.indexOf) 
			{
				var new_object = [];
				for (var i = 0; i < object.length; i++) 
					new_object.push(this._formatResponse(object[i]));
			}
			else 
			{
				var new_object = {};
				for (var key in object) 
					new_object[key] = this._formatResponse(object[key]);
			}
		}
		else if (object != '') 
		{
			new_object = object;
		}
		return new_object;
	},
}
